/***************************************************************************
   dw-output.c  -  Output data to eps and txt files.
   Modified output-eps.c from Autotrace by Martin Weber.
                             -------------------
    begin                : Thu Sep 5 2002
    copyright            : (C) 2002 by Adam Tofilski
    email                : rotofils@cyf-kr.edu.pl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "types.h"
#include "spline.h"
#include "color.h"
#include "dw-output.h"
#include "xstd.h"
#include "autotrace.h"
#include <time.h>
#include <string.h>

static at_string now (void);

#define SIGN(x) ((x) > 0 ? 1 : (x) < 0 ? -1 : 0)
#define ROUND(x) ((int) ((int) (x) + .5 * SIGN (x)))

/* Output macros.  */

/* This should be used for outputting a string S on a line by itself.  */
#define OUT_LINE(s)							\
  fprintf (ps_file, "%s\n", s)

/* These output their arguments, preceded by the indentation.  */
#define OUT1(s, e)							\
  fprintf (ps_file, s, e)

#define OUT2(s, e1, e2)							\
  fprintf (ps_file, s, e1, e2)

#define OUT3(s, e1, e2, e3)							\
  fprintf (ps_file, s, e1, e2, e3)

#define OUT4(s, e1, e2, e3, e4)						\
  fprintf (ps_file, s, e1, e2, e3, e4)

#define OUT5(s, e1, e2, e3, e4, e5)					\
  fprintf (ps_file, s, e1, e2, e3, e4, e5)

/* These macros just output their arguments.  */
#define OUT_STRING(s)	fprintf (ps_file, "%s", s)
#define OUT_REAL(r)	fprintf (ps_file, r == (ROUND (r = ROUND((at_real)6.0*r)/(at_real)6.0))				\
                                  ? "%.0f " : "%.3f ", r)

/* For a PostScript command with two real arguments, e.g., lineto.  OP
   should be a constant string.  */
#define OUT_COMMAND2(first, second, op)					\
  do									\
    {									\
      OUT_REAL (first);							\
      OUT_REAL (second);						\
      OUT_STRING (op "\n");						\
    }									\
  while (0)

/* For a PostScript command with six real arguments, e.g., curveto.
   Again, OP should be a constant string.  */
#define OUT_COMMAND6(first, second, third, fourth, fifth, sixth, op)	\
  do									\
    {									\
      OUT_REAL (first);							\
      OUT_REAL (second);						\
      OUT_STRING (" ");							\
      OUT_REAL (third);							\
      OUT_REAL (fourth);						\
      OUT_STRING (" ");							\
      OUT_REAL (fifth);							\
      OUT_REAL (sixth);							\
      OUT_STRING (" " op " \n");						\
    }									\
  while (0)

/* This should be called before the others in this file.  It opens the
   output file `OUTPUT_NAME.ps', and writes some preliminary boilerplate. */

static void
output_eps_header(FILE* ps_file, at_string name,
                  int llx, int lly, int urx, int ury)
{
  at_string time;

  OUT_LINE ("%!PS-Adobe-3.0 EPSF-3.0");
  OUT_LINE ("%%%%Creator: DrawWing 0.11");
  OUT1 ("%%%%Title: %s\n", name);
  OUT1 ("%%%%CreationDate: %s\n", time = now ());
  OUT4 ("%%%%BoundingBox: %d %d %d %d\n", llx, lly, urx, ury);
  OUT_LINE ("%%DocumentData: Clean7Bit");
  OUT_LINE ("%%EndComments");

  free (time);
  /* Prolog to define Illustrator commands.*/

  OUT_LINE ("%%BeginProlog");
  OUT_LINE ("/bd { bind def } bind def");
  OUT_LINE ("/m { moveto } bd");
  OUT_LINE ("/l { lineto } bd");
  OUT_LINE ("/c { curveto } bd");
  OUT_LINE ("/S { stroke } bd");
  OUT_LINE ("/k { setcmykcolor } bd"); /* must symbol k for CorelDraw 3/4 */
  OUT_LINE ("/K { k } bd");
  OUT_LINE ("%%EndProlog");
  OUT_LINE ("%%BeginSetup"); /* needed for CorelDraw 3/4 */
  OUT_LINE ("%%EndSetup");  /* needed for CorelDraw 3/4 */
}

/* This outputs the PostScript code which produces the shape in
   SHAPE.  */

static void
out_splines (FILE * ps_file, at_splines_type shape)
{
  unsigned this_list;
  //started from 1 to omit picture border
  for (this_list = 1; this_list < SPLINE_LIST_ARRAY_LENGTH (shape);
       this_list++)
    {
      unsigned this_spline;

      spline_list_type list = SPLINE_LIST_ARRAY_ELT (shape, this_list);
      spline_type first = SPLINE_LIST_ELT (list, 0);

      OUT_COMMAND2 (START_POINT (first).x, START_POINT (first).y, "m");

      for (this_spline = 0; this_spline < SPLINE_LIST_LENGTH (list);
           this_spline++)
      {
         spline_type s = SPLINE_LIST_ELT (list, this_spline);

         if (SPLINE_DEGREE (s) == LINEARTYPE)
           OUT_COMMAND2 (END_POINT (s).x, END_POINT (s).y, "l");
         else
           OUT_COMMAND6 (CONTROL1 (s).x, CONTROL1 (s).y,
                         CONTROL2 (s).x, CONTROL2 (s).y,
                         END_POINT (s).x, END_POINT (s).y,
                         "c");
      }
      OUT_LINE ("S");
    }
}


void
dw_output_txt_writer (FILE* txt_file, at_string name,
                      int dpi, at_splines_type splines)
{
  unsigned this_list, point, index, counter;
  at_real_coord *temp;
  at_real_coord **junctions;

  XMALLOC(junctions,
          SPLINE_LIST_ARRAY_LENGTH (splines)*2*sizeof(at_real_coord *));

  for (this_list = 0; this_list < SPLINE_LIST_ARRAY_LENGTH (splines);
       this_list++)
  {
      junctions[this_list]= &splines.data[this_list].data[0].v[0];
      junctions[this_list+SPLINE_LIST_ARRAY_LENGTH (splines)]
        =&splines.data[this_list].data[splines.data[this_list].length-1].v[3];
  }

  for (point = 0; point < 2*SPLINE_LIST_ARRAY_LENGTH (splines)-1;
       point++)
  {
    for (index = point+1; index < 2*SPLINE_LIST_ARRAY_LENGTH (splines);
       index++)
    {

      if(junctions[point]->x > junctions[index]->x){
        temp=junctions[index];
        junctions[index]=junctions[point];
        junctions[point]=temp;
      }
      else if(junctions[point]->x == junctions[index]->x)
        if(junctions[point]->y > junctions[index]->y){
          temp=junctions[index];
          junctions[index]=junctions[point];
          junctions[point]=temp;
        }
    }
  }
  counter=0;
  for (point = 0; point < 2*SPLINE_LIST_ARRAY_LENGTH (splines)-1;
       point++)
  {
      if((junctions[point]->x==junctions[point+1]->x)
         && (junctions[point]->y==junctions[point+1]->y))
        counter++;
      else{
        if(counter>1)
          fprintf(txt_file, "%8.2f %8.2f\n", junctions[point]->x, junctions[point]->y);
        counter=0;
      }
  }
  fclose (txt_file);
  free(junctions);
}


void
dw_output_eps_writer(FILE* ps_file, at_string name,
                     int dpi, at_splines_type shape)
{
  int urx = shape.width;
  int ury = shape.height;
  at_real scale = shape.wing_width/100.0;
  unsigned margin = 72;              //1 inch
  //    unsigned page_w = 595;
  unsigned page_h = 842;
  unsigned narrow_border = 5;
  unsigned wide_border = 35;
  unsigned bbllx = margin;           //left margin
//  unsigned bblly = page_h-margin-100-3*narrow_border-wide_border;//lower margin
  unsigned bblly = page_h-margin-ury/scale-3*narrow_border-wide_border;//lower margin
  unsigned bburx = margin+urx/scale+3*narrow_border+wide_border; //right margin 
  unsigned bbury = page_h-margin;           //upper margin
  unsigned bb_w = bburx - bbllx;
  unsigned bb_h = bbury - bblly;

  output_eps_header(ps_file, name, bbllx, bblly, bburx, bbury);

  fprintf (ps_file,"0 0 0 1 K\n");
  fprintf (ps_file,"/Helvetica findfont 10 scalefont setfont\n");
  fprintf (ps_file,"0.25 setlinewidth\n");
  fprintf (ps_file,"/str 10 string def\n");
  fprintf (ps_file,"%d %d translate\n", bbllx+wide_border+narrow_border,
                                        bblly+wide_border+narrow_border);
  fprintf (ps_file,"%d %d %d %d rectstroke\n", -narrow_border, -narrow_border,
                                               bb_w-wide_border-narrow_border,
                                               bb_h-wide_border-narrow_border);
  fprintf (ps_file, "%d %d m\n", -narrow_border, -narrow_border);
  fprintf (ps_file, "0 50 %d {dup %d m 0 -5 rlineto 0 -12 rmoveto str cvs show} for\n",
                    bb_w-wide_border-2*narrow_border, -narrow_border);
  fprintf (ps_file,"0 0 m\n");
  fprintf (ps_file,"0 50 %d {dup %d exch m -5 0 rlineto -20 0 rmoveto str cvs show} for\n",
                   bb_h-wide_border-narrow_border, -narrow_border);
  if(dpi)
  {
    fprintf (ps_file,"%d %d m\n", 0, -2);
    fprintf (ps_file,"%f 0 rlineto\n", dpi/(25.6*scale));
    fprintf (ps_file,"%d %d m\n", 0, 0);
    fprintf (ps_file,"(1 mm) show\n");
  }

  out_splines(ps_file, shape);
//  OUT_LINE ("S");

  OUT_LINE ("%%Trailer");
  OUT_LINE ("%%EOF");
}


static at_string
now (void)
{
  at_string time_string;
  time_t t = time (0);

  XMALLOC (time_string, 26); /* not 25 ! */
  strcpy (time_string, ctime (&t));
  time_string[24] = 0; /* No newline. */

  return time_string;
}

