/***************************************************************************
   main.c  -  Convert wing image to Bezier curves.
   Modified main.c from Autotrace by Martin Weber.

                             -------------------
    begin                : Thu Sep 5 2002
    copyright            : (C) 2002 by Adam Tofilski
    email                : rotofils@cyf-kr.edu.pl
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "cmdline.h"
#include "filename.h"
#include "xstd.h"
#include "atou.h"
#include "logreport.h"
#include "drawwing.h"
#include "dw-output.h"

#include <assert.h>

/* Pointers to functions based on input format.  (-input-format)  */
static at_input_read_func input_reader = NULL;

/* Return NAME with any leading path stripped off.  This returns a
   pointer into NAME.  For example, `basename ("/foo/bar.baz")'
   returns "bar.baz".  */
static char * get_basename (char * name);

/* The name of the file we're going to write.  (-output-file) */
static char * output_name = (char *)"";

/* The name of the file we're going to write.  (-output-file) */
static char * bgd_name = (char *)"";

/* Whether to print version information */
static at_bool printed_version;

/* Whether to write a log file */
static at_bool logging = false;

/* image dpi used in mif backend. */
static int dpi = 0;

static char * read_command_line (int, char * [], at_fitting_opts_type *);

void input_list_formats(FILE * file);


int
main (int argc, char * argv[])
{
  at_fitting_opts_type * fitting_opts;
  char * input_name, * input_rootname, * logfile_name=NULL;
  at_splines_type * splines;
  at_bitmap_type * bitmap;
  at_bitmap_type * bgd;
  FILE *output_file;

  fitting_opts = at_fitting_opts_new ();
  fitting_opts->outline = 0;
  fitting_opts->venation = 0;
  fitting_opts->shortest_vein = 10;

  input_name = read_command_line (argc, argv, fitting_opts);

  input_rootname = remove_suffix (get_basename (input_name));
  if (input_rootname == NULL)
   FATAL1 ("Not a valid inputname %s\n", input_name);

  if (logging)
    log_file = xfopen (logfile_name = extend_filename (input_rootname, "log"), "w");

  if (logging)
    free (logfile_name);

  /* Set input_reader if it is not set in command line args */
  if (!input_reader)
    input_reader = at_input_get_handler (input_name);

  /* Open the main input file.  */
  if (input_reader != NULL)
    bitmap = at_bitmap_read(input_reader, input_name);
  else
    FATAL ("Unsupported inputformat\n");

  /* Open the bacground file.  */
  if (strcmp (bgd_name, ""))
  {
    bgd = at_bitmap_read(input_reader, bgd_name);
    subtruct(bitmap, *bgd);
    at_bitmap_free (bgd);
  }

  LOG1("bitmap.height %d\n", bitmap->height);
  LOG1("bitmap.width %d\n", bitmap->width);
  LOG1("bitmap.np %d\n", bitmap->np);
  if(bitmap->np > 1)
    FATAL ("Not 8 bit per pixel bitmap\n");

  splines = dw_splines(bitmap, fitting_opts);

  /* Open output file */
  output_name = extend_filename (input_rootname, "eps");
  output_file = xfopen(output_name, "wb");
  dw_output_eps_writer(output_file, input_rootname, dpi, *splines);
  fclose (output_file);

  output_name = extend_filename (input_rootname, "txt");
  output_file = xfopen(output_name, "w");
  dw_output_txt_writer(output_file, " ", dpi, *splines);
  fclose (output_file);

  free(output_name);
  free(input_rootname);
  at_splines_free (splines);
  at_bitmap_free (bitmap);
  at_fitting_opts_free(fitting_opts);

  return 0;
}


/* Reading the options.  */

#define USAGE "<input_name.bmp> should be a windows bitmap.\n\
Options:\n"\
  GETOPT_USAGE                        \
"outline <unsigned>: threshold value for outline discrimination. \n\
venation <unsigned>: threshold value for venation discrimination. \n\
shortest-vein <unsigned>: delete shorter veins. \n\
background <input_name.bmp>: use background image to remove uneven lighting\n\
dpi <unsigned>: dots per inch value in the input image, affects scaling\n\
  of output image\n\
help: print this message.\n\
log: write detailed progress reports to <input_name>.log.\n\
version: print the version number of this program.\n\
"

/* We return the name of the image to process.  */
static char *
read_command_line (int argc, char * argv[],
                   at_fitting_opts_type * fitting_opts)
{
  int g;   /* `getopt' return code.  */
  int option_index;
  struct option long_options[]
    = {
        { "dpi",                   1, 0, 0 },
        { "help",                  0, 0, 0 },
        { "log",                   0, (int *) &logging, 1 },
        { "version",               0, (int *) &printed_version, 1 },
        { "outline",               1, 0, 0 },
        { "venation",              1, 0, 0 },
        { "shortest-vein",         1, 0, 0 },
        { "background",            1, 0, 0 },
        { 0, 0, 0, 0 } };

  while (true)
    {
      g = getopt_long_only (argc, argv, "", long_options, &option_index);

      if (g == EOF)
        break;

      if (g == '?')
        exit (1);  /* Unknown option.  */

      assert (g == 0); /* We have no short option names.  */

      if (ARGUMENT_IS ("dpi"))
        dpi = atou (optarg);

      else if (ARGUMENT_IS ("help"))
        {
          fprintf (stderr, "Usage: %s [options] <input_name.bmp>.\n", argv[0]);
          fprintf (stderr, USAGE);
          fprintf (stderr,
                   "\nYou can get the source code of DrawWing from \n%s\n",
                   dw_home_site());
          exit (0);
        }

      else if (ARGUMENT_IS ("version"))
        printf ("%s\n", dw_version()   );

      else if (ARGUMENT_IS ("outline"))
        fitting_opts->outline = atou (optarg);

      else if (ARGUMENT_IS ("venation"))
        fitting_opts->venation = atou (optarg);

      else if (ARGUMENT_IS ("shortest-vein"))
        fitting_opts->shortest_vein = atou (optarg);

      else if (ARGUMENT_IS ("background"))
        bgd_name = optarg;

      /* Else it was just a flag; getopt has already done the assignment.  */
    }
  /* Just wanted to know the version number?  */
  if (printed_version && optind == argc) exit (0);

  /* Exactly one (non-empty) argument left?  */
  if (optind + 1 == argc && *argv[optind] != 0)
    {
      return (argv[optind]);
    }
  else
    {
      fprintf (stderr, "Usage: %s [options] <image_name.bmp>.\n", argv[0]);
      fprintf (stderr, "(%s.)\n", optind == argc ? "Missing <image_name.bmp>"
                                  : "Too many <image_name.bmp>s");
      fputs ("For more information, use ``-help''.\n", stderr);
      exit (1);
    }
  return NULL; /* stop warnings */
}

/* Return NAME with any leading path stripped off.  This returns a
   pointer into NAME.  For example, `basename ("/foo/bar.baz")'
   returns "bar.baz".  */

static char *
get_basename (char * name)
{
#ifdef WIN32
  char * base = strrchr (name, '\\');
#else
  char * base = strrchr (name, '/');
#endif
  return base ? base + 1 : name;
}

void
input_list_formats(FILE * file)
{
  char ** list = at_input_list_new ();
  char ** tmp;

  char * suffix;
  char * descr;
  tmp = list;
  while (*list)
    {
      suffix = *list++;
      descr = *list++;
      fprintf(file, "%5s %s\n", suffix, descr);
    }

  at_input_list_free(tmp);
}

